/*
    tstore - an archive format for archiving tar archives
    Copyright (C) 2005  Kasper Dupont

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    with permission from the author any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    $Id: tself.c,v 1.6 2005/06/21 12:57:34 kasperd Exp $
*/

#include "tstore.h"
#include "test.h"

#define DATA(_i) (data[idx[(_i)]])

int main(int argc, char ** argv)
{
  struct store *store;
  FILE *hash_file;
  int levels;
  int64_t filesize;
  int i,N,rfd;
  struct timeval tv;

  uint8_t *hash_values;
  uint8_t (*data)[512];
  int *idx;

  if (argc < 3) {
    fprintf(stderr,"Uage: %s <store> <filenames>\n",argv[0]);
    return 1;
  }

  for (i=2;i<argc;++i)
    if (strlen(argv[i])>TS_MAX_NAME_LEN) {
      fprintf(stderr,"%s: Name too long\n",argv[2]);
      return 1;
    }

  rfd=open("/dev/urandom",O_RDONLY);
  if (rfd==-1) errstop("/dev/urandom");

  store=open_store(argv[1],O_RDWR,CHECK_LOCK);
  if (!store) errstop(argv[1]);

  hash_file=fopen(store->h32name,"r+");
  if (hash_file) {
    if (fread(hash_key,1,16,hash_file)!=16) errstop(store->h32name);
  }

  for (levels=1,i=0;i<4;++i) {
    N=levels+argc-1;
    filesize=512*(store->store_size+N);
    levels=filelevels(filesize);
  }

  data=myalloc(N*512);
  idx=myalloc(N*sizeof(int));
  hash_values=myalloc(N*4);

  for (i=0;i<N;++i) idx[i]=i;
  for (i=1;i<N;++i) {
    uint64_t v;
    int j,t;
    if (myread(rfd,&v,8)!=8) errstop("read(\"/dev/urandom\")");
    j=1+v%i;

    t     =idx[i];
    idx[i]=idx[j];
    idx[j]=t;
  }

  pread(store->fd,data,512,0);

  for (i=0;i<levels;++i) {
    char *buf=DATA(i+1);
    int j,step;
    uint64_t L=idx[i]+store->store_size;
    PUT_BNUM(buf,L);
    for (step=1,j=0;j<i;++j) step*=102;
    for (j=1;j<102;++j)
      if (j*step<(filesize>>9))
	PUT_BNUM(buf+5*j,j*step+i*0x4000000000ULL);
  }

  if(gettimeofday(&tv,NULL)) errstop("gettimeofday");

  for (i=2;i<argc;++i) {
    uint8_t *fileheader=DATA(levels+i-1);

    if (i==2) {
      PUT_BNUM(fileheader+NEXTPTR_OFFSET,GET_BNUM(DATA(0)+NEXTPTR_OFFSET));
      PUT_BNUM(fileheader+STORESIZE_OFFSET,store->store_size);
    } else {
      PUT_BNUM(fileheader+NEXTPTR_OFFSET,idx[levels+i-2]+store->store_size);
      PUT_BNUM(fileheader+STORESIZE_OFFSET,store->store_size+N);
    }
    PUT_BNUM(fileheader+TREEPTR_OFFSET,idx[levels]+store->store_size);
    PUT_SIZE(fileheader+FILESIZE_OFFSET,filesize);
    PUT_NTIM(fileheader+NANOSEC_OFFSET,tv.tv_usec*1000);
    PUT_TIME(fileheader+SECONDS_OFFSET,tv.tv_sec);
    strcpy(fileheader+FILENAME_OFFSET,argv[i]);
  }

  PUT_BNUM(DATA(0)+NEXTPTR_OFFSET,idx[N-1]+store->store_size);
  PUT_BNUM(DATA(0)+STORESIZE_OFFSET,store->store_size+N);

  if (lock_store(store)) errstop(store->name);

  pwrite(store->fd,data,N*512,store->store_size*512);
  pwrite(store->fd,DATA(0)+STORESIZE_OFFSET,10,STORESIZE_OFFSET);

  if (hash_file) {
    for (i=0;i<N;++i) {
      uint8_t buffer[HBS];
      memcpy(buffer+16,data[i],512);
      hash_buf(buffer,512);
      memcpy(hash_values+4*i,buffer,4);
    }
    if(fseek(hash_file,(store->store_size+3)*4,SEEK_SET)||
       (fwrite(hash_values,1,4*N,hash_file)!=4*N)||
       fflush(hash_file)) errstop(store->h32name);
  }

  close_store(store);

  return 0;
}
